/*
 * **    spasm - The spasm anti-spam milter
 * **    Copyright (C) 2002-2005 Sean Hanson <aradia@theasylum.org>
 * **
 * **    This program is free software; you can redistribute it and/or modify
 * **    it under the terms of the GNU General Public License as published by
 * **    the Free Software Foundation; either version 2 of the License, or
 * **    (at your option) any later version.
 * **
 * **    This program is distributed in the hope that it will be useful,
 * **    but WITHOUT ANY WARRANTY; without even the implied warranty of
 * **    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * **    GNU General Public License for more details.
 * **
 * **    You should have received a copy of the GNU General Public License
 * **    along with this program; if not, write to the Free Software
 * **    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * */

#include <arpa/inet.h>
#include <curses.h>
#include <dirent.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>

#include "spasm.h"
#include "dbfunc.h"

extern struct spasmConfigStruct *spasm_config;

/* Low-level QDBM functions */

/* qdbm_lookup_entry() - lookup a qdbm hash entry, return the value */
char*
qdbm_lookup_entry(char *dbfile, char *key)
{
    DEPOT *db;
    char *buf;

    if(dbfile == NULL || key == NULL) return 0;

    if((db = dpopen(dbfile, DP_OREADER, -1)) == NULL) {
	if(dpecode != 8) spasm_error(LOG_ERR, "(qdbm_lookup_entry) Error opening database %s: %s", dbfile, dperrmsg(dpecode));
	return NULL;
    }
    if((buf = dpget(db, key, strlen(key), 0, SPASM_DBT_MAX, NULL)) == NULL) {
        if(dpecode != DP_ENOITEM) spasm_error(LOG_ERR, "(qdbm_lookup_entry) Error getting value from database %s: %s", dbfile, dperrmsg(dpecode));
	dpclose(db);
        return NULL;

    }
    dpclose(db);
    return buf;
}

/* qdbm_insert_entry() - insert a qdbm hash entry */
void
qdbm_insert_entry(char *dbfile, char *key, char *value)
{     
    DEPOT *db;

    if(dbfile == NULL || key == NULL || value == NULL) return;

    if((db = dpopen(dbfile, DP_OWRITER|DP_OCREAT, -1)) == NULL) {
        spasm_error(LOG_ERR, "(qdbm_insert_entry) Error opening database %s: %s (%m)", dbfile, dperrmsg(dpecode));
        return;
    }
    if(dpput(db, key, strlen(key), value, strlen(value), DP_DOVER) == 0) {
        spasm_error(LOG_ERR, "(qdbm_insert_entry) Error adding %s/%s to %s: %s (%m)", key, value, dbfile, dperrmsg(dpecode));
    }
    if(dpclose(db) == 0) {
        spasm_error(LOG_ERR, "(qdbm_insert_entry) Error closing database %s: %s (%m)", dbfile, dperrmsg(dpecode));
    }
    return;
}

/* qdbm_delete_entry() - delete a qdbm hash entry */
void
qdbm_delete_entry(char *dbfile, char *key)
{     
    DEPOT *db;

    if(dbfile == NULL || key == NULL) return;

    if((db = dpopen(dbfile, DP_OWRITER|DP_OCREAT, -1)) == NULL) {
        spasm_error(LOG_ERR, "(qdbm_delete_entry) Error opening database %s: %s", dbfile, dperrmsg(dpecode));
        return;
    }
    if(dpout(db, key, strlen(key)) == 0) {
        spasm_error(LOG_ERR, "(qdbm_delete_entry) Error deleting %s from %s: %s", key, dbfile, dperrmsg(dpecode));
    }
    if(dpclose(db) == 0) {
        spasm_error(LOG_ERR, "(qdbm_delete_entry) Error closing database %s: %s", dbfile, dperrmsg(dpecode));
    }
    return;
}

/* qdbm_dump_entries() - dump spasm_list of entries in qdbm database */
struct spasm_list*
qdbm_dump_entries(char *dbfile, int vals)
{   
    struct spasm_list *cur = NULL;
    struct spasm_list *first = NULL;
    DEPOT *db;
    char *key;
    char buf[SPASM_DBT_MAX+1];
    int ksize;
    int vsize;

    if(dbfile == NULL) return NULL;

    if((db = dpopen(dbfile, DP_OREADER, -1)) == NULL) {
        if(dpecode != 8) spasm_error(LOG_ERR, "(qdbm_dump_entries) Error opening database %s: %s", dbfile, dperrmsg(dpecode));
        return NULL;
    }
    if(dpiterinit(db) == 0) {
        spasm_error(LOG_ERR, "(qdbm_dump_entries) Error initialising iterator for %s: %s", dbfile, dperrmsg(dpecode));
	dpclose(db);
        return NULL;
    }
    while((key = dpiternext(db, &ksize)) != NULL) {
	if((cur = calloc(sizeof(struct spasm_list), 1)) != NULL) {
	    if(ksize <= SPASM_DBT_MAX) {
		strncpy(cur->entry, key, ksize);
		if(vals) {
		    memset(buf, 0, sizeof(buf));
		    if((vsize = dpgetwb(db, key, strlen(key), 0, SPASM_DBT_MAX, buf)) == -1)
			spasm_error(LOG_ERR, "(qdbm_dump_entries) Error getting value from database %s: %s", dbfile, dperrmsg(dpecode));
		    else
			strncpy(cur->strval, buf, vsize);
		}
		if(first != NULL) cur->next = first;
		first = cur;
	    } else spasm_error(LOG_ERR, "(qdbm_dump_entries) Entry is too large");
	} else spasm_error(LOG_ERR, "(qdbm_dump_entries) Error allocating memory");
	free(key);
    }
    dpclose(db);
    return first;
}

/* qdbm_set_db() - set db with spasm_list of entries  */
int
qdbm_set_db(char *dbfile, struct spasm_list *entries)
{
    DEPOT *db;
    struct spasm_list *first;

    if(dbfile == NULL || entries == NULL) return 0;

    if((db = dpopen(dbfile, DP_OWRITER|DP_OCREAT|DP_OTRUNC, -1)) == NULL) {
        spasm_error(LOG_ERR, "(qdbm_set_db) Error creating db: %s", dperrmsg(dpecode));
        return 0;
    }
    while(entries != NULL) {
        char *value;
        if(*entries->strval != '\0') value = entries->strval;
        else value = "1";
        if(dpput(db, entries->entry, strlen(entries->entry), value, strlen(value), DP_DOVER) == 0)
            spasm_error(LOG_ERR, "(qdbm_set_db) Error adding %s to %s: %s", entries->entry, dbfile, dperrmsg(dpecode));
        first = entries->next;
        free(entries);
        entries = first;
    }
    dpclose(db);
    return 1;
}


/* DB Abstraction Functions */

/* spasm_verify_filters() - verify that a user has filters set */
int
spasm_verify_filters(char *username)
{
    DEPOT *db;
    char dbfile[FILENAME_MAX+1];

    if(username == NULL) return 0;

    USR_STRCAT(dbfile, username, "filters");
    if((db = dpopen(dbfile, DP_OREADER, -1)) != NULL) {
	dpclose(db);
	return 1;
    }

    return 0;
}

/* spasm_lookup_filter() - look up a filter */
int
spasm_lookup_filter(char *username, char *entry)
{
    char dbfile[FILENAME_MAX+1];
    char *domain, *result = NULL;
    int rv = 0;

    if(username == NULL || entry == NULL) return 0;

    if(spasm_verify_filters(username)) {
	USR_STRCAT(dbfile, username, "filters");
	result = qdbm_lookup_entry(dbfile, entry);
	if(result != NULL) rv = atoi(result);
    } else if((domain = strchr(username, '@')) != NULL) {
	domain++;
	USR_STRCAT(dbfile, domain, "filters");
	result = qdbm_lookup_entry(dbfile, entry);
	if(result != NULL) rv = atoi(result);
    }
    if(result != NULL) free(result);

    return rv;
}

/* spasm_get_filters() - get filters */
struct spasm_list*
spasm_get_filters(char *username)
{   
    char dbfile[FILENAME_MAX+1];
    struct spasm_list *entries = NULL;
    char *domain;
    
    if(username == NULL) return NULL;

    if(spasm_verify_filters(username)) {
        USR_STRCAT(dbfile, username, "filters");
        entries = qdbm_dump_entries(dbfile, 1);
    } else if((domain = strchr(username, '@')) != NULL) {
        domain++;
        USR_STRCAT(dbfile, domain, "filters");
        entries = qdbm_dump_entries(dbfile, 1);
    }

    return entries;
}

/* spasm_set_filter() - set a filter */
void
spasm_set_filter(char *username, char *entry, int value)
{
    char buf[128];
    char dbfile[FILENAME_MAX+1];
    if(username == NULL || entry == NULL) return;
    USR_STRCAT(dbfile, username, "filters");
    snprintf(buf, sizeof(buf), "%d", value);
    qdbm_insert_entry(dbfile, entry, buf);
    return;
}

/* spasm_lookup_db() - look up a system database entry */
int 
spasm_lookup_db(char *database, char *entry)
{   
    char dbfile[FILENAME_MAX+1];
    char *result;
    if(database == NULL || entry == NULL) return 0;
    SYS_STRCAT(dbfile, database);
    result = qdbm_lookup_entry(dbfile, entry);
    if(result == NULL) return 0;
    free(result);
    return 1;
}   

/* spasm_get_db() - get system database entries */
struct spasm_list*
spasm_get_db(char *database)
{
    char dbfile[FILENAME_MAX+1];
    struct spasm_list *entries = NULL;
    if(database == NULL) return NULL;
    SYS_STRCAT(dbfile, database);
    entries = qdbm_dump_entries(dbfile, 1);
    return entries;
}

/* spasm_insert_db() - add a system database entry */
void
spasm_insert_db(char *database, char *entry, char *val)
{
    char dbfile[FILENAME_MAX+1];
    if(database == NULL || entry == NULL || val == NULL) return;
    SYS_STRCAT(dbfile, database);
    qdbm_insert_entry(dbfile, entry, val);
    return;
}

/* spasm_delete_db() - delete a system database entry */
void
spasm_delete_db(char *database, char *entry)
{   
    char dbfile[FILENAME_MAX+1];
    if(database == NULL || entry == NULL) return;
    SYS_STRCAT(dbfile, database);
    qdbm_delete_entry(dbfile, entry);
    return;
}

/* returns:  1 - all is well, 0 - failure */
int
spasm_set_db(char *database, struct spasm_list *entries)
{
    char dbfile[FILENAME_MAX+1];
    if(database == NULL || entries == NULL) return 0;
    SYS_STRCAT(dbfile, database);
    return qdbm_set_db(dbfile, entries);
}

/* spasm_lookup_bl() - look up a blacklist entry */
int
spasm_lookup_bl(char *username, char *entry)
{
    char dbfile[FILENAME_MAX+1];
    char *result;
    if(username == NULL || entry == NULL) return 0;
    USR_STRCAT(dbfile, username, "blacklist");
    result = qdbm_lookup_entry(dbfile, entry);
    if(result == NULL) return 0;
    free(result);
    return 1;
}

/* spasm_get_bl() - get blacklist entries */
struct spasm_list*
spasm_get_bl(char *username)
{
    char dbfile[FILENAME_MAX+1];
    struct spasm_list *entries = NULL;
    if(username == NULL) return NULL;
    USR_STRCAT(dbfile, username, "blacklist");
    entries = qdbm_dump_entries(dbfile, 1);
    return entries;
}

/* spasm_insert_bl() - add a blacklist entry */
void
spasm_insert_bl(char *username, char *entry, char *val)
{
    char dbfile[FILENAME_MAX+1];
    if(username == NULL || entry == NULL || val == NULL) return;
    USR_STRCAT(dbfile, username, "blacklist");
    qdbm_insert_entry(dbfile, entry, val);
    return;
}

/* spasm_delete_bl() - delete a blacklist entry */
void
spasm_delete_bl(char *username, char *entry)
{
    char dbfile[FILENAME_MAX+1];
    if(username == NULL || entry == NULL) return;
    USR_STRCAT(dbfile, username, "blacklist");
    qdbm_delete_entry(dbfile, entry);
    return;
}

/* returns:  1 - all is well, 0 - failure */
int
spasm_set_bl(char *username, struct spasm_list *entries)
{   
    char dbfile[FILENAME_MAX+1];
    if(username == NULL || entries == NULL) return 0;
    USR_STRCAT(dbfile, username, "blacklist");
    return qdbm_set_db(dbfile, entries);
}

/* spasm_lookup_wl() - look up whitelist entries */
int
spasm_lookup_wl(char *username, char *entry)
{
    char dbfile[FILENAME_MAX+1];
    char *domain, *result = NULL;
    if(username == NULL || entry == NULL) return 0;
    USR_STRCAT(dbfile, username, "whitelist");
    if((result = qdbm_lookup_entry(dbfile, entry)) == NULL) {
	if((domain = strchr(username, '@')) != NULL) {
	    domain++;
	    USR_STRCAT(dbfile, domain, "whitelist");
	    result = qdbm_lookup_entry(dbfile, entry);
	}
    }
    if(result == NULL) return 0;
    free(result);
    return 1;
}

/* spasm_get_wl() - get whitelist entries */
struct spasm_list*
spasm_get_wl(char *username)
{
    char dbfile[FILENAME_MAX+1];
    struct spasm_list *entries = NULL;
    if(username == NULL) return NULL;
    USR_STRCAT(dbfile, username, "whitelist");
    entries = qdbm_dump_entries(dbfile, 1);
    return entries;
}

/* spasm_insert_wl() - add a whitelist entry */
void
spasm_insert_wl(char *username, char *entry, char *val)
{
    char dbfile[FILENAME_MAX+1];
    if(username == NULL || entry == NULL || val == NULL) return;
    USR_STRCAT(dbfile, username, "whitelist");
    qdbm_insert_entry(dbfile, entry, val);
    return;
}

/* spasm_delete_wl() - delete a whitelist entry */
void
spasm_delete_wl(char *username, char *entry)
{
    char dbfile[FILENAME_MAX+1];
    if(username == NULL || entry == NULL) return;
    USR_STRCAT(dbfile, username, "whitelist");
    qdbm_delete_entry(dbfile, entry);
    return;
}

/* returns:  1 - all is well, 0 - failure */
int
spasm_set_wl(char *username, struct spasm_list *entries)
{   
    char dbfile[FILENAME_MAX+1];
    if(username == NULL || entries == NULL) return 0;
    USR_STRCAT(dbfile, username, "whitelist");
    return qdbm_set_db(dbfile, entries);
}   

/* spasm_lookup_prefs() - look up a user preference */
char*
spasm_lookup_prefs(char *username, char *entry)
{
    char dbfile[FILENAME_MAX+1];
    char *domain, *value = NULL;

    if(username == NULL || entry == NULL) return 0;
    USR_STRCAT(dbfile, username, "prefs");
    if((value = qdbm_lookup_entry(dbfile, entry)) == NULL) {
	if((domain = strchr(username, '@')) != NULL) {
	    domain++;
	    USR_STRCAT(dbfile, domain, "prefs");
	    value = qdbm_lookup_entry(dbfile, entry);
	}
    }
    return value;
}

/* spasm_set_prefs() - insert a user preference */
void
spasm_set_prefs(char *username, char *entry, char *value)
{
    char dbfile[FILENAME_MAX+1];
    if(username == NULL || entry == NULL || value == NULL) return;
    USR_STRCAT(dbfile, username, "prefs");
    qdbm_insert_entry(dbfile, entry, value);
    return;
}

/* spasm_del_prefs() - delete a user preference */
void
spasm_del_prefs(char *username, char *entry)
{
    char dbfile[FILENAME_MAX+1];
    if(username == NULL || entry == NULL) return;
    USR_STRCAT(dbfile, username, "prefs");
    qdbm_delete_entry(dbfile, entry);
    return;
}

/* spasm_get_log() - get a log entry */
struct spasm_list*
spasm_get_log(char *username, char *key, int all)
{
    struct spasm_list *entries = NULL;
    char dbname[FILENAME_MAX+1];
    char *ptr;

    if(username == NULL || (key == NULL && all == 0)) return NULL;

    if(key != NULL) {	/* Get one log entry */
	USR_STRCAT(dbname, username, "logs");
	if((entries = calloc(sizeof(struct spasm_list), 1)) == NULL) {
	    spasm_error(LOG_ERR, "(spasm_get_log) Error allocating memory");
	    return NULL;
	}
	if((ptr = qdbm_lookup_entry(dbname, key)) != NULL) {
	    strncpy(entries->entry, key, sizeof(entries->entry));
	    strncpy(entries->strval, ptr, sizeof(entries->strval));
	    free(ptr);
	    if((ptr = strchr(entries->strval, '\n')) != NULL) *ptr = '\0';
	} else {
	    spasm_error(LOG_ERR, "(spasm_get_log) Error getting log entry");
	    free(entries);
	    return NULL;
	}
    } else if(all) {	/* Get all log entries */
	USR_STRCAT(dbname, username, "logs");
	entries = qdbm_dump_entries(dbname, 1);
    }

    return entries;
}

/* spasm_delete_log() - delete a log entry */
void
spasm_delete_log(char *username, char *key, int all)
{
    char dbfile[FILENAME_MAX+1];

    if(username == NULL || (key == NULL && all == 0)) return;

    if(key != NULL) {	/* Delete one log entry */
	USR_STRCAT(dbfile, username, "logs");
	qdbm_delete_entry(dbfile, key);
    } else if(all) {	/* Delete all log entries */
	USR_STRCAT(dbfile, username, "logs");
	unlink(dbfile);
    }
    return;
}

/* spasm_insert_log() - add a log entry */
void
spasm_insert_log(char *username, char *entry)
{
    char key[SPASM_DBT_MAX+1];
    char dbfile[FILENAME_MAX+1];
    char *ptr;

    if(username == NULL || entry == NULL) return;

    USR_STRCAT(dbfile, username, "logs");
    if((ptr = strchr(entry, ':')) == NULL) {
	spasm_error(LOG_ERR, "(spasm_insert_log) Error parsing timestamp");
	return;
    }

    /* Take the key (timestamp) directly from the log entry */
    *ptr = '\0';
    snprintf(key, sizeof(key), "%s", entry);
    *ptr = ':';

    qdbm_insert_entry(dbfile, key, entry);

    return;
}

/* spasm_update_stats() - update statistics */
void
spasm_update_stats(char *key)
{
    char dbfile[FILENAME_MAX+1];
    char *value = NULL;
    char buf[SPASM_DBT_MAX+1];
    struct tm tms;
    time_t timep;
    int i = 0;

    if(key == NULL) return;

    timep = time(NULL);
    if(localtime_r(&timep, &tms) == NULL) {
	spasm_error(LOG_ERR, "(spasm_update_stats) Error reading localtime");
	return;
    }
    snprintf(dbfile, sizeof(dbfile), "%s/stats/%4d-%02d-%02d.db", spasm_config->spasmroot, tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday);
    if((value = qdbm_lookup_entry(dbfile, key)) != NULL) {
	i = atoi(value);
	free(value);
    }
    i++;
    snprintf(buf, sizeof(buf), "%d", i);
    qdbm_insert_entry(dbfile, key, buf);
    return;
}

/* End DB Abstraction Functions */

/* Create user directory */
void
spasm_create_user(char *username)
{
    char dirname[FILENAME_MAX+1];

    snprintf(dirname, sizeof(dirname), "%s/users/%s", spasm_config->spasmroot, username);
    mkdir(dirname, 0755);
    snprintf(dirname, sizeof(dirname), "%s/users/%s/quarantine", spasm_config->spasmroot, username);
    mkdir(dirname, 0755);
    return;
}
